
;+
;NAME:
;     IDEPROJECT
;PURPOSE:
;     Warp an image so that it is deprojected from its true disk position to 
;     disk center.
;CATEGORY:
;CALLING SEQUENCE:
;     image = ideproject(image,point)
;INPUTS:
;     image = image array to warp (can be a data cube, in which case
;              point must have one element per image)
;     point = structure giving pointing information.  The structure must 
;             contain at least the following tags:
;                B0 = Solar B0 angle in radians
;                P  = Solar P angle in radians.  Set this to zero if the image
;                     has solar north up already.
;                LAT = Latitude of the center of the FOV in radians
;                CMD = CMD of the center of the FOV in radians
;                PIX_SIZE = [x_pixel_size,y_pixel_size] in arcseconds
;                RADIUS = radius of sun in arcseconds
;             Point can also be a mag structure from bfits.pro
;             For SXT PFI images, use get_point_str.pro to get the point
;             structure.
;OPTIONAL INPUT PARAMETERS:
;KEYWORD PARAMETERS
;     bytscale = if set, byte scale the output image so that the lowest 15
;                colors are reserved for contours etc.  The largest value
;                will depend on how many colors are available (!d.table_size)
;     missing = the image value to use for blank regions in the transformation.
;     negative = if set, return a negative image (only if /bytscale is set)
;     /cubic = Set this to use cubic interpolation in poly_2d.
;     /nointerp = Set this to avoid interpolation in poly_2d.
;OUTPUTS:
;     image = deprojected image array.  The spatial scale of the output array
;             is the same as the input array, but the dimension of the output
;             is larger due to the deprojection.  I.e. each pixel in the
;             output array has the same size, in arc seconds, as the input
;             array, in arcseconds. 
;COMMON BLOCKS:
;SIDE EFFECTS:
;RESTRICTIONS:
;     Assumes that the image is of something on the photosphere.  Since a 
;     rotation is done, this will not work for X-ray images ... without 
;     knowing the height of the x-ray structures, the deprojection can not
;     be done correctly.
;
;     The deprojection does not work well if some of the pixels are off the
;     limb. 
;PROCEDURE:
;EXAMPLE:
;     For SXT PFI images (white light only! See restrictions):
;         point = get_point_str(index)
;         newdata = ideproject(data,point)
;MODIFICATION HISTORY:
;
;       Nov, 1992  T. Metcalf
;       1997-Oct-30 T. Metcalf - Improved the algorithm to compute the lat
;                              and CMD at every point in the image
;                              independently.  This accounts for solar
;                              curvature.  To take advantage of this
;                              the PIX_SIZE and RADIUS elements are 
;                              required in the point structure.  If they
;                              are not present, the old behavior of the
;                              routine is used.
;                              - Also changed the loop to assume that there
;                              is a seperate point structure for each image
;                              in the data cube.
;      1999-Jun-25 T. Metcalf  - Added cubic and nointerp keywords.
;-

function ideproject,dimage,point,negative=negative,missing=missing, $
                    bytscale=bytscale,cubic=cubic,nointerp=nointerp, $ 
                    b_pxd=b_pxd,b_pyd=b_pyd

     nx_image = n_elements(dimage(*,0,0))
     ny_image = n_elements(dimage(0,*,0))
     xcorners = [0,0,nx_image-1,nx_image-1]
     ycorners = [0,ny_image-1,0,ny_image-1]
     
     if tag_index(point,'LAT') LT 0 then begin
        pindex = tag_index(point,'POINT')
        if pindex LT 0 then message,'Point structure is incomplete'
        upoint = point.(pindex)
     endif $
     else upoint=point

     dsize = size(dimage)
     if dsize(0) EQ 3 then nimage = dsize(3) $
     else if dsize(0) EQ 2 then nimage = 1 $
     else message,'Image array does not have the correct number of dimensions'

     if nimage NE n_elements(upoint) then begin
        if n_elements(upoint) eq 1 then begin
           upoint = replicate(upoint,nimage) 
           message,/info,'WARNING: point structure has only one element, replicating to match the number of images in the data cube.'
        endif else $
           message,'The point structure must have the same number of elements as the number of images in the data cube'
     endif

     if tag_index(upoint,'B0') LT 0 OR tag_index(upoint,'P') LT 0 OR $
        tag_index(upoint,'LAT') LT 0 OR tag_index(upoint,'CMD') LT 0 then $
        message,'Point structure is incomplete'

     newpoint = upoint

     for i=0L,nimage-1L do begin

        ;message,/info,'Image '+string(i)

        ; Set up the deprojection transformation

        b0 = double(upoint(i).b0)
        p  = double(upoint(i).p)
        if tag_index(upoint,'PIX_SIZE') GE 0 AND $
           tag_index(upoint,'RADIUS') GE 0 then begin
           latlon = get_latlon(nx=nx_image,ny=ny_image,point=upoint(i))
           bc = double(latlon.latitude) * !dtor
           lc = double(latlon.cmd) * !dtor
        endif else begin
           bc = double(upoint(i).lat)
           lc = double(upoint(i).cmd)
           message,/info,'WARNING: PIX_SIZE and/or RADIUS not set.  Treating image as a plane.'
        endelse

        cxx = - SIN(b0)*SIN(p)*SIN(lc) + COS(p)*COS(lc)
        cxy = - SIN(bc)*( SIN(b0)*SIN(p)*COS(lc) + COS(p)*SIN(lc) ) $
              - COS(bc)*COS(b0)*SIN(p) 
        cyx = + SIN(b0)*COS(p)*SIN(lc) + SIN(p)*COS(lc)
        cyy = + SIN(bc)*( SIN(b0)*COS(p)*COS(lc) - SIN(p)*SIN(lc) ) $
              + COS(bc)*COS(b0)*COS(p)
 
        determ = cxx*cyy-cyx*cxy  ; Get image to helio transformation
        icxx = cyy/determ         ; The determinent is double precision
        icxy = -cxy/determ
        icyx = -cyx/determ
        icyy = cxx/determ
 
        dXdxi = icxx
        dYdxi = icyx
        dXdyi = icxy
        dYdyi = icyy
 
        if tag_index(upoint,'PIX_SIZE') LT 0 then begin
           pix_size = [1.d0,1.d0]
        endif else begin
           pix_size = double(upoint(i).pix_size)
           if n_elements(pix_size) EQ 1 then $
              pix_size = [pix_size(0),pix_size(0)]
        endelse
        mpix_size = min(pix_size)
 
        x = (lindgen(nx_image,ny_image) MOD nx_image)*(pix_size(0)/mpix_size); image pixel coordinates
        y = (lindgen(nx_image,ny_image) / nx_image)*(pix_size(1)/mpix_size) 
        b_pxd = dXdxi*x + dXdyi*y           ; Heliographic pixel coordinates
        b_pyd = dYdxi*x + dYdyi*y


        nx = fix((max(b_pxd)-min(b_pxd)+1))
        ny = fix((max(b_pyd)-min(b_pyd)+1))

        if i EQ 0 then begin 
           newimage=make_array(nx,ny,nimage,type=dsize(dsize(0)+1))
           newnx = nx
           newny = ny
        endif 
        xoffset = (newnx - nx)/2.0
        yoffset = (newny - ny)/2.0

        ; p,q will be the forward poly_2d transformation 
        if n_elements(missing) NE 1 then $
           missing =(max(dimage(*,*,i))+min(dimage(*,*,i)))/2.
        polywarp,x-xoffset,y-yoffset,b_pxd-min(b_pxd), $
                 b_pyd-min(b_pyd),3,p,q
        if keyword_set(cubic) or NOT keyword_set(nointerp) then begin
           useinterp =1+keyword_set(cubic) 
           if keyword_set(cubic) then cval = -0.5 ; otherwise leave undefined
        endif else begin
           useinterp = 0
        endelse
        newimage(*,*,i) = poly_2d(dimage(*,*,i),p,q,useinterp, $
                                  newnx,newny,missing=missing,cubic=cval)

        if keyword_set(bytscale) then begin
           ; The bytscl call ensures that the image will not use
           ; the first 15 colors which will be reserved for contours.
           newimage(*,*,i)= (15+bytscl(newimage(*,*,i),top=(!d.table_size-1-15)))
           if keyword_set(negative) then begin
              newimage(*,*,i) = max(newimage(*,*,i))   $
                                + min(newimage(*,*,i)) $
                                - newimage(*,*,i)
           endif
        endif
     endfor

     return,newimage

end
